#!/usr/bin/perl

# Copyright (C) 2008  Anders Kaseorg <andersk@mit.edu>,
#                     Jeffrey Brian Arnold <jbarnold@mit.edu>,
#                     Tim Abbott <tabbott@mit.edu>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street - Fifth Floor, Boston, MA
# 02110-1301, USA.

use strict;
use warnings;
use lib '/usr/share/ksplice';
use Ksplice;

$Verbose::level = $ENV{KSPLICE_VERBOSE} if (defined $ENV{KSPLICE_VERBOSE});

sub empty_diff {
	my ($out) = @_;
	my ($obj) = $out =~ /^(.*)\.KSPLICE$/ or die;
	unlink "$obj.KSPLICE_primary" if (-e "$obj.KSPLICE_primary");
	unlink "$obj.KSPLICE_helper" if (-e "$obj.KSPLICE_helper");
	open OUT, '>', "$out.tmp";
	close OUT;
	rename "$out.tmp", $out;
}

sub do_snap {
	my ($out) = @_;
	my ($obj) = $out =~ /^(.*)\.KSPLICE$/ or die;
	die if (!-e $obj);
	unlink "$obj.KSPLICE_pre" if (-e "$obj.KSPLICE_pre");
	empty_diff($out);
}

sub do_diff {
	my ($out) = @_;
	my ($obj) = $out =~ /^(.*)\.KSPLICE$/ or die;
	my $obj_old = "$obj.KSPLICE_pre";
	die if (!-e $obj);
	die "Patch creates new object $obj" if (!-e $obj_old);
	if (system('cmp', '-s', '--', $obj_old, $obj) == 0) {
		unlink $obj_old;
		return empty_diff($out);
	}

	runval("$libexecdir/ksplice-objmanip", $obj, "$obj.KSPLICE_primary", "keep-primary", "$obj.KSPLICE_pre", $ENV{KSPLICE_KID});

	open OUT, '>', "$out.tmp";
	print OUT "1\n";
	close OUT;
	rename "$out.tmp", $out;

	runval("$libexecdir/ksplice-objmanip", $obj_old, "$obj.KSPLICE_helper", "keep-helper");
}

sub do_helper {
	my ($out) = @_;
	my ($obj) = $out =~ /^(.*)\.KSPLICE_helper$/ or die;
	my $obj_old = "$obj.KSPLICE_pre";
	-e $obj_old or $obj_old = $obj;
	runval("$libexecdir/ksplice-objmanip", $obj_old, "$obj.KSPLICE_helper", "keep-helper");
}

sub link_objs {
	my ($out, @ins) = @_;
	if (@ins == 0) {
		runval(shellwords($ENV{AR}), "rcs", $out);
	} elsif (@ins == 1) {
		copy @ins, $out;
	} else {
		runval(shellwords($ENV{LD}), "-r", "-o", $out, @ins);
	}
}

sub do_combine {
	my ($out, @ins) = @_;
	my @primaries;
	my @helpers;
	foreach my $in (@ins) {
		if (my ($obj) = $in =~ /^(.*)\.KSPLICE$/) {
			next if (!-s $in);
			push @primaries, "$obj.KSPLICE_primary";
			push @helpers, "$obj.KSPLICE_helper";
		} elsif (($obj) = $in =~ /^(.*)\.KSPLICE_helper$/) {
			push @helpers, "$obj.KSPLICE_helper"
			    unless (@helpers && $helpers[$#helpers] eq "$obj.KSPLICE_helper");
		} else {
			die;
		}
	}

	return empty_diff($out) unless (@helpers);

	my ($obj) = $out =~ /^(.*)\.KSPLICE$/ or die;
	link_objs("$obj.KSPLICE_primary", @primaries);
	link_objs("$obj.KSPLICE_helper", @helpers);

	open OUT, '>', "$out.tmp";
	print OUT "1\n";
	close OUT;
	rename "$out.tmp", $out;
}

sub do_finalize {
	my ($in, $out) = @_;
	runval("$libexecdir/ksplice-objmanip", $in, $out, "finalize");
}

sub do_rmsyms {
	my ($in, $out, @rmsyms) = @_;
	runval_in("@rmsyms\n", "$libexecdir/ksplice-objmanip", $in, $out, "rmsyms");
}

sub do_system_map_lookup {
	my ($symarg) = @_;
	open(SYMS, "<", "$ENV{KSPLICE_CONFIG_DIR}/System.map") or die;
	my $line;
	while (defined($line = <SYMS>)) {
		my ($addr, $type, $sym, $mod) = split(/\s+/, $line);
		if ($sym eq $symarg) { print $addr; last; }
	}
	close(SYMS);
}

my %handlers = (
	'snap' => \&do_snap,
	'diff' => \&do_diff,
	'helper' => \&do_helper,
	'combine' => \&do_combine,
	'finalize' => \&do_finalize,
	'rmsyms' => \&do_rmsyms,
	'system_map_lookup' => \&do_system_map_lookup,
);

my ($cmd, @args) = @ARGV;
if (exists $handlers{$cmd}) {
	my $handler = $handlers{$cmd};
	&$handler(@args);
} else {
	print "Usage: ksplice-obj.pl ", join('|', keys %handlers), " ...\n";
	exit(1);
};
